from stable_baselines.common.vec_env.base_vec_env import VecEnv


class SBVecEnvWrapper(VecEnv):
    def __init__(self, venv):
        self.venv = venv
        self.num_envs = venv.num_envs
        self.observation_space = venv.observation_space
        self.action_space = venv.action_space

    def reset(self):
        return self.venv.reset()

    def step_async(self, actions):
        self.venv.step_async(actions)

    def step_wait(self):
        return self.venv.step_wait()

    def step(self, actions):
        return self.venv.step(actions)

    def close(self):
        del self.venv

    def seed(self, seed=None):
        self.venv.seed(seed)

    def get_attr(self, attr_name, indices=None):
        raise NotImplementedError()

    def set_attr(self, attr_name, value, indices=None):
        raise NotImplementedError()

    def env_method(self, method_name, *method_args, indices=None, **method_kwargs):
        raise NotImplementedError()



from pettingzoo.utils.wrappers import OrderEnforcingWrapper as PettingzooWrap


class BaseWrapper(PettingzooWrap):
    def __init__(self, env):
        """
        Creates a wrapper around `env`. Extend this class to create changes to the space.
        """
        super().__init__(env)

        self._check_wrapper_params()

        self._modify_spaces()

    def _check_wrapper_params(self):
        pass

    def _modify_spaces(self):
        pass

    def _modify_action(self, agent, action):
        raise NotImplementedError()

    def _modify_observation(self, agent, observation):
        raise NotImplementedError()

    def _update_step(self, agent):
        pass

    def reset(self):
        super().reset()
        self._update_step(self.agent_selection)

    def observe(self, agent):
        obs = super().observe(agent)
        observation = self._modify_observation(agent, obs)
        return observation

    def step(self, action):
        agent = self.env.agent_selection
        if not self.dones[agent]:
            action = self._modify_action(agent, action)

        super().step(action)

        self._update_step(self.agent_selection)



from .utils.base_modifier import BaseModifier
from .utils.shared_wrapper_util import shared_wrapper
import gym


def sticky_actions_v0(env, repeat_action_probability):
    assert 0 <= repeat_action_probability < 1

    class StickyActionsModifier(BaseModifier):
        def __init__(self):
            self.seed(None)

        def reset(self):
            self.old_action = None

        def seed(self, seed):
            self.np_random, _ = gym.utils.seeding.np_random(seed)

        def modify_action(self, action):
            if self.old_action is not None and self.np_random.uniform() < repeat_action_probability:
                action = self.old_action
            self.old_action = action
            return action

    return shared_wrapper(env, StickyActionsModifier)



import numpy as np
from pettingzoo.test import api_test, seed_test, parallel_test
from pettingzoo.mpe import simple_push_v2, simple_world_comm_v2
from pettingzoo.butterfly import knights_archers_zombies_v7, prison_v3
from pettingzoo.classic import chess_v3

import supersuit
import pytest
from supersuit import (
    frame_stack_v1,
    pad_action_space_v0,
    frame_skip_v0,
    sticky_actions_v0,
    dtype_v0
)


def test_pettingzoo_frame_stack():
    _env = simple_push_v2.env()
    wrapped_env = frame_stack_v1(_env)
    api_test(wrapped_env)


def test_pettingzoo_frame_skip():
    env = simple_push_v2.raw_env(max_cycles=100)
    env = frame_skip_v0(env, 3)
    env.reset()
    x = 0
    for agent in env.agent_iter(25):
        assert env.env.steps == (x // 2) * 3
        action = env.action_spaces[env.agent_selection].sample()
        next_obs = env.step(action)
        x += 1


def test_pettingzoo_pad_action_space():
    _env = simple_world_comm_v2.env()
    wrapped_env = pad_action_space_v0(_env)
    api_test(wrapped_env)
    seed_test(lambda: sticky_actions_v0(simple_world_comm_v2.env(), 0.5), 100)


def test_pettingzoo_parallel_env():
    _env = simple_world_comm_v2.parallel_env()
    wrapped_env = pad_action_space_v0(_env)
    parallel_test.parallel_api_test(wrapped_env)


wrappers = [
    supersuit.color_reduction_v0(knights_archers_zombies_v7.env(), "R"),
    supersuit.resize_v0(dtype_v0(knights_archers_zombies_v7.env(), np.uint8), x_size=5, y_size=10),
    supersuit.resize_v0(dtype_v0(knights_archers_zombies_v7.env(), np.uint8), x_size=5, y_size=10, linear_interp=True),
    supersuit.dtype_v0(knights_archers_zombies_v7.env(), np.int32),
    supersuit.flatten_v0(knights_archers_zombies_v7.env()),
    supersuit.reshape_v0(knights_archers_zombies_v7.env(), (512 * 512, 3)),
    supersuit.normalize_obs_v0(dtype_v0(knights_archers_zombies_v7.env(), np.float32), env_min=-1, env_max=5.0),
    supersuit.frame_stack_v1(knights_archers_zombies_v7.env(), 8),
    supersuit.pad_observations_v0(knights_archers_zombies_v7.env()),
    supersuit.pad_action_space_v0(knights_archers_zombies_v7.env()),
    supersuit.black_death_v1(knights_archers_zombies_v7.env()),
    supersuit.agent_indicator_v0(knights_archers_zombies_v7.env(), True),
    supersuit.agent_indicator_v0(knights_archers_zombies_v7.env(), False),
    supersuit.reward_lambda_v0(knights_archers_zombies_v7.env(), lambda x: x / 10),
    supersuit.clip_reward_v0(knights_archers_zombies_v7.env()),
    supersuit.clip_actions_v0(prison_v3.env(continuous=True)),
    supersuit.scale_actions_v0(prison_v3.env(continuous=True), 0.5),
    supersuit.nan_noop_v0(knights_archers_zombies_v7.env(), 0),
    supersuit.nan_zeros_v0(knights_archers_zombies_v7.env()),
    supersuit.nan_zeros_v0(prison_v3.env(continuous=True)),
    supersuit.nan_random_v0(chess_v3.env()),
    supersuit.nan_random_v0(knights_archers_zombies_v7.env()),
    supersuit.frame_skip_v0(knights_archers_zombies_v7.env(), 4),
    supersuit.sticky_actions_v0(knights_archers_zombies_v7.env(), 0.75),
    supersuit.delay_observations_v0(knights_archers_zombies_v7.env(), 3),
    supersuit.max_observation_v0(knights_archers_zombies_v7.env(), 3),
]


@pytest.mark.parametrize("env", wrappers)
def test_pettingzoo_aec_api(env):
    api_test(env)


parallel_wrappers = [
    supersuit.frame_stack_v1(knights_archers_zombies_v7.parallel_env(), 8),
    supersuit.reward_lambda_v0(knights_archers_zombies_v7.parallel_env(), lambda x: x / 10),
    supersuit.delay_observations_v0(knights_archers_zombies_v7.parallel_env(), 3),
    supersuit.color_reduction_v0(knights_archers_zombies_v7.parallel_env(), "R"),
    supersuit.frame_skip_v0(knights_archers_zombies_v7.parallel_env(), 4),
    supersuit.max_observation_v0(knights_archers_zombies_v7.parallel_env(), 4),
    supersuit.black_death_v1(knights_archers_zombies_v7.parallel_env()),
]


@pytest.mark.parametrize("env", parallel_wrappers)
def test_pettingzoo_parallel_api(env):
    parallel_test.parallel_api_test(env)



from .utils.base_modifier import BaseModifier
from .utils.shared_wrapper_util import shared_wrapper
import warnings
from supersuit.lambda_wrappers import action_lambda_v1
import numpy as np
import gym


def nan_random_v0(env):
    class NanRandomModifier(BaseModifier):
        def __init__(self):
            super().__init__()
            self.seed()

        def modify_action(self, action):
            if action is not None and np.isnan(action).any():
                obs = self.cur_obs
                if isinstance(obs, dict) and 'action mask' in obs:
                    warnings.warn("[WARNING]: Step received an NaN action {}. Environment is {}. Taking a random action from 'action mask'.".format(action, self))
                    action = self.np_random.choice(np.flatnonzero(obs['action_mask']))
                else:
                    warnings.warn("[WARNING]: Step received an NaN action {}. Environment is {}. Taking a random action.".format(action, self))
                    action = self.action_space.sample()
            return action

        def seed(self, seed=None):
            self.np_random, seed = gym.utils.seeding.np_random(seed)

    return shared_wrapper(env, NanRandomModifier)


def nan_noop_v0(env, no_op_action):
    def on_action(action, action_space):
        if np.isnan(action).any():
            warnings.warn("[WARNING]: Step received an NaN action {}. Evironment is {}. Taking no operation action.".format(action, env))
            action = no_op_action
        return action
    return action_lambda_v1(env, on_action, lambda act_space: act_space)


def nan_zeros_v0(env):
    def on_action(action, action_space):
        if np.isnan(action).any():
            warnings.warn("[WARNING]: Step received an NaN action {}. Environment is {}. Taking the all zeroes action.".format(action, env))
            action = np.zeros_like(action)
        return action
    return action_lambda_v1(env, on_action, lambda act_space: act_space)



from pettingzoo.utils.env import AECEnv, ParallelEnv
from supersuit.utils.action_transforms import homogenize_ops
from supersuit import observation_lambda_v0, action_lambda_v1


def pad_action_space_v0(env):
    assert isinstance(env, AECEnv) or isinstance(env, ParallelEnv), "pad_action_space_v0 only accepts an AECEnv or ParallelEnv"
    homogenize_ops.check_homogenize_spaces(list(env.action_spaces.values()))
    padded_space = homogenize_ops.homogenize_spaces(list(env.action_spaces.values()))
    return action_lambda_v1(env,
        lambda action, act_space: homogenize_ops.dehomogenize_actions(act_space, action),
        lambda act_space: padded_space)


def pad_observations_v0(env):
    assert isinstance(env, AECEnv) or isinstance(env, ParallelEnv), "pad_observations_v0 only accepts an AECEnv or ParallelEnv"
    spaces = list(env.observation_spaces.values())
    homogenize_ops.check_homogenize_spaces(spaces)
    padded_space = homogenize_ops.homogenize_spaces(spaces)
    return observation_lambda_v0(env,
        lambda obs, obs_space: homogenize_ops.homogenize_observations(obs_space, obs),
        lambda obs_space: padded_space)



from pettingzoo.utils import ParallelEnv
from gym.spaces import Box, Discrete
import numpy as np
import supersuit


class DummyParEnv(ParallelEnv):
    metadata = {"render.modes": ["human"]}

    def __init__(self, observations, observation_spaces, action_spaces):
        super().__init__()
        self._observations = observations
        self.observation_spaces = observation_spaces

        self.agents = [x for x in observation_spaces.keys()]
        self.possible_agents = self.agents
        self.agent_selection = self.agents[0]
        self.action_spaces = action_spaces

        self.rewards = {a: 1 for a in self.agents}
        self.dones = {a: False for a in self.agents}
        self.infos = {a: {} for a in self.agents}

    def step(self, actions):
        for agent, action in actions.items():
            assert action in self.action_spaces[agent]
        return self._observations, self.rewards, self.dones, self.infos

    def reset(self):
        return self._observations

    def close(self):
        pass


base_obs = {"a{}".format(idx): np.zeros([8, 8, 3], dtype=np.float32) + np.arange(3) + idx for idx in range(2)}
base_obs_space = {"a{}".format(idx): Box(low=np.float32(0.0), high=np.float32(10.0), shape=[8, 8, 3]) for idx in range(2)}
base_act_spaces = {"a{}".format(idx): Discrete(5) for idx in range(2)}


def test_basic():
    env = DummyParEnv(base_obs, base_obs_space, base_act_spaces)
    env = supersuit.delay_observations_v0(env, 4)
    env = supersuit.dtype_v0(env, np.uint8)
    orig_obs = env.reset()
    for i in range(10):
        action = {agent: env.action_spaces[agent].sample() for agent in env.agents}
        obs, rew, done, info = env.step(action)



import multiprocessing as mp
import numpy as np


class SharedArray:
    def __init__(self, shape, dtype):
        self.shared_arr = mp.Array(np.ctypeslib.as_ctypes_type(dtype), int(np.prod(shape)), lock=False)
        self.dtype = dtype
        self.shape = shape
        self._set_np_arr()

    def _set_np_arr(self):
        self.np_arr = np.frombuffer(self.shared_arr, dtype=self.dtype).reshape(self.shape)

    def __getstate__(self):
        return (self.shared_arr, self.dtype, self.shape)

    def __setstate__(self, state):
        self.shared_arr = state
        self._set_np_arr()



class DeprecatedWrapper(ImportError):
    pass


class Deprecated:
    def __init__(self, wrapper_name, orig_version, new_version):
        self.name = wrapper_name
        self.old_version, self.new_version = orig_version, new_version

    def __call__(self, env, *args, **kwargs):
        raise DeprecatedWrapper(f"{self.name}_{self.old_version} is now Deprecated, use {self.name}_{self.new_version} instead")



from collections import deque
import numpy as np
from functools import reduce


class Accumulator:
    def __init__(self, obs_space, memory, reduction):
        self.memory = memory
        self._obs_buffer = deque()
        self.reduction = reduction
        self.maxed_val = None

    def add(self, in_obs):
        self._obs_buffer.append(np.copy(in_obs))
        if len(self._obs_buffer) > self.memory:
            self._obs_buffer.popleft()
        self.maxed_val = None

    def get(self):
        if self.maxed_val is None:
            self.maxed_val = reduce(self.reduction, (self._obs_buffer))
        return self.maxed_val



from .utils.base_modifier import BaseModifier
from .utils.shared_wrapper_util import shared_wrapper
from supersuit.utils.obs_delay import Delayer


def delay_observations_v0(env, delay):
    class DelayObsModifier(BaseModifier):
        def reset(self):
            self.delayer = Delayer(self.observation_space, delay)

        def modify_obs(self, obs):
            obs = self.delayer.add(obs)
            return BaseModifier.modify_obs(self, obs)

    return shared_wrapper(env, DelayObsModifier)



from collections import deque
import numpy as np


class Delayer:
    def __init__(self, obs_space, delay):
        self.delay = delay
        self.obs_queue = deque()
        self.obs_space = obs_space

    def add(self, in_obs):
        self.obs_queue.append(in_obs)
        if len(self.obs_queue) > self.delay:
            return self.obs_queue.popleft()
        else:
            return np.zeros_like(in_obs)



import numpy as np
from .single_vec_env import SingleVecEnv
import gym.vector


def transpose(ll):
    return [[ll[i][j] for i in range(len(ll))] for j in range(len(ll[0]))]


class ConcatVecEnv(gym.vector.VectorEnv):
    def __init__(self, vec_env_fns, obs_space=None, act_space=None):
        self.vec_envs = vec_envs = [vec_env_fn() for vec_env_fn in vec_env_fns]
        for i in range(len(vec_envs)):
            if not hasattr(vec_envs[i], "num_envs"):
                vec_envs[i] = SingleVecEnv([lambda: vec_envs[i]])
        self.metadata = self.vec_envs[0].metadata
        self.observation_space = vec_envs[0].observation_space
        self.action_space = vec_envs[0].action_space
        tot_num_envs = sum(env.num_envs for env in vec_envs)
        self.num_envs = tot_num_envs
        self.obs_buffer = np.empty((self.num_envs,) + self.observation_space.shape, dtype=self.observation_space.dtype)

    def concat_obs(self, obs_list):
        idx = 0
        for venv, obs in zip(self.vec_envs, obs_list):
            endidx = idx + venv.num_envs
            self.obs_buffer[idx:endidx] = obs
            idx = endidx
        return self.obs_buffer

    def seed(self, seed=None):
        if seed is None:
            for env in self.vec_envs:
                env.seed(None)
        else:
            for env in self.vec_envs:
                env.seed(seed)
                seed += env.num_envs

    def reset(self):
        return self.concat_obs([vec_env.reset() for vec_env in self.vec_envs])

    def step_async(self, actions):
        self._saved_actions = actions

    def step_wait(self):
        return self.step(self._saved_actions)

    def step(self, actions):
        data = []
        idx = 0
        for venv in self.vec_envs:
            data.append(venv.step(actions[idx: idx + venv.num_envs]))
            idx += venv.num_envs
        observations, rewards, dones, infos = transpose(data)
        observations = self.concat_obs(observations)
        rewards = np.concatenate(rewards, axis=0)
        dones = np.concatenate(dones, axis=0)
        infos = sum(infos, [])
        return observations, rewards, dones, infos

    def render(self, mode="human"):
        return self.vec_envs[0].render(mode)

    def close(self):
        for vec_env in self.vec_envs:
            vec_env.close()

    def env_is_wrapped(self, wrapper_class):
        return sum([sub_venv.env_is_wrapped(wrapper_class) for sub_venv in self.vec_envs], [])



from .concat_vec_env import ConcatVecEnv
from .multiproc_vec import ProcConcatVec


class call_wrap:
    def __init__(self, fn, data):
        self.fn = fn
        self.data = data

    def __call__(self, *args):
        return self.fn(self.data)


def MakeCPUAsyncConstructor(max_num_cpus):
    if max_num_cpus == 0 or max_num_cpus == 1:
        return ConcatVecEnv
    else:

        def constructor(env_fn_list, obs_space, act_space):
            example_env = env_fn_list[0]()
            envs_per_env = getattr(example_env, "num_envs", 1)

            num_fns = len(env_fn_list)
            envs_per_cpu = (num_fns + max_num_cpus - 1) // max_num_cpus
            alloced_num_cpus = (num_fns + envs_per_cpu - 1) // envs_per_cpu

            env_cpu_div = []
            num_envs_alloced = 0
            while num_envs_alloced < num_fns:
                start_idx = num_envs_alloced
                end_idx = min(num_fns, start_idx + envs_per_cpu)
                env_cpu_div.append(env_fn_list[start_idx:end_idx])
                num_envs_alloced = end_idx

            assert alloced_num_cpus == len(env_cpu_div)

            cat_env_fns = [call_wrap(ConcatVecEnv, env_fns) for env_fns in env_cpu_div]
            return ProcConcatVec(cat_env_fns, obs_space, act_space, num_fns * envs_per_env, example_env.metadata)

        return constructor



def check_transform_frameskip(frame_skip):
    if isinstance(frame_skip, tuple) and len(frame_skip) == 2 and isinstance(frame_skip[0], int) and isinstance(frame_skip[1], int) and 1 <= frame_skip[0] <= frame_skip[1]:
        return frame_skip
    elif isinstance(frame_skip, int):
        return (frame_skip, frame_skip)
    else:
        assert False, "frame_skip must be an int or a tuple of two ints, where the first values is at least one and not greater than the second"



class VectorAECEnv:
    def reset(self):
        """
        resets all environments
        """

    def seed(self, seed=None):
        """
        seeds all environments
        """

    def observe(self, agent):
        """
        returns observation for agent from all environments (if agent is alive, else all zeros)
        """

    def last(self, observe=True):
        """
        returns list of observations, rewards, dones, env_dones, passes, infos

        each of the following is a list over environments that holds the value for the current agent (env.agent_selection)

        dones: are True when the current agent is done
        env_dones: is True when all agents are done, and the environment will reset
        passes: is true when the agent is not stepping this turn (because it is dead or not currently stepping for some other reason)
        infos: list of infos for the agent
        """

    def step(self, actions, observe=True):
        """
        steps the current agent with the following actions.
        Unlike a regular AECEnv, the actions cannot be None
        """

    def agent_iter(self, max_iter):
        """
        Unlike aec agent_iter, this does not stop on environment done. Instead,
        vector environment resets speciic envs when done.

        Instead, just continues until max_iter is reached.
        """
        return AECIterable(self, max_iter)


class AECIterable:
    def __init__(self, env, max_iter):
        self.env = env
        self.max_iter = max_iter

    def __iter__(self):
        return AECIterator(self.env, self.max_iter)


class AECIterator:
    def __init__(self, env, max_iter):
        self.env = env
        self.iters_til_term = max_iter
        self.env._is_iterating = True

    def __next__(self):
        if self.iters_til_term <= 0:
            raise StopIteration
        self.iters_til_term -= 1
        return self.env.agent_selection



import gym
import numpy as np


class SpaceWrapper:
    def __init__(self, space):
        if isinstance(space, gym.spaces.Discrete):
            self.shape = ()
            self.dtype = np.dtype(np.int64)
        elif isinstance(space, gym.spaces.Box):
            self.shape = space.shape
            self.dtype = np.dtype(space.dtype)
        else:
            assert False, "ProcVectorEnv only support Box and Discrete types"



import numpy as np
import gym


class SingleVecEnv:
    def __init__(self, gym_env_fns, *args):
        assert len(gym_env_fns) == 1
        self.gym_env = gym_env_fns[0]()
        self.observation_space = self.gym_env.observation_space
        self.action_space = self.gym_env.action_space
        self.num_envs = 1
        self.metadata = self.gym_env.metadata

    def reset(self):
        return np.expand_dims(self.gym_env.reset(), 0)

    def step_async(self, actions):
        self._saved_actions = actions

    def step_wait(self):
        return self.step(self._saved_actions)

    def seed(self, seed=None):
        self.gym_env.seed(seed)

    def render(self, mode="human"):
        return self.gym_env.render(mode)

    def close(self):
        self.gym_env.close()

    def step(self, actions):
        observations, reward, done, info = self.gym_env.step(actions[0])
        if done:
            observations = self.gym_env.reset()
        observations = np.expand_dims(observations, 0)
        rewards = np.array([reward], dtype=np.float32)
        dones = np.array([done], dtype=np.uint8)
        infos = [info]
        return observations, rewards, dones, infos

    def env_is_wrapped(self, wrapper_class):
        env_tmp = self.gym_env
        while isinstance(env_tmp, gym.Wrapper):
            if isinstance(env_tmp, wrapper_class):
                return [True]
            env_tmp = env_tmp.env
        return [False]



import numpy as np
from . import convert_box


def check_param(obs_space, new_dtype):
    np.dtype(new_dtype)  # type argument must be convertable to a numpy dtype


def change_obs_space(obs_space, param):
    return convert_box(lambda obs: change_observation(obs, obs_space, param), obs_space)


def change_observation(obs, obs_space, new_dtype):
    obs = obs.astype(new_dtype)
    return obs



from .generic_wrappers import * # NOQA
from .lambda_wrappers import action_lambda_v1, observation_lambda_v0, reward_lambda_v0 # NOQA
from .multiagent_wrappers import agent_indicator_v0, black_death_v1, black_death_v0, \
    pad_action_space_v0, pad_observations_v0 # NOQA
from supersuit.generic_wrappers import frame_skip_v0, color_reduction_v0, resize_v0, dtype_v0, \
    flatten_v0, reshape_v0, normalize_obs_v0, clip_actions_v0, clip_reward_v0, \
    delay_observations_v0, frame_stack_v1, frame_stack_v0, max_observation_v0, \
    sticky_actions_v0 # NOQA

from .vector.vector_constructors import gym_vec_env_v0, stable_baselines_vec_env_v0, \
    stable_baselines3_vec_env_v0, concat_vec_envs_v0, pettingzoo_env_to_vec_env_v0 # NOQA

from .aec_vector import vectorize_aec_env_v0 # NOQA


__version__ = "3.0.1"


import warnings
warnings.warn("You're using SuperSuit 3.0, released 7/7/21. The entire codebase has been rewritten or refactored as part of this release. While we've tested it thoroughly, please ensure everything you're doing still works properly and report any issues at https://github.com/PettingZoo-Team/SuperSuit. This warning will be removed 2 months after release.")



from supersuit.utils.base_aec_wrapper import PettingzooWrap
from supersuit.utils.wrapper_chooser import WrapperChooser
import gym


class aec_reward_lambda(PettingzooWrap):
    def __init__(self, env, change_reward_fn):
        assert callable(change_reward_fn), "change_reward_fn needs to be a function. It is {}".format(change_reward_fn)
        self._change_reward_fn = change_reward_fn

        super().__init__(env)

    def _check_wrapper_params(self):
        pass

    def _modify_spaces(self):
        pass

    def reset(self):
        super().reset()
        self.rewards = {agent: self._change_reward_fn(reward) for agent, reward in self.rewards.items()}
        self.__cumulative_rewards = {a: 0 for a in self.agents}
        self._accumulate_rewards()

    def step(self, action):
        agent = self.env.agent_selection
        super().step(action)
        self.rewards = {agent: self._change_reward_fn(reward) for agent, reward in self.rewards.items()}
        self.__cumulative_rewards[agent] = 0
        self._cumulative_rewards = self.__cumulative_rewards
        self._accumulate_rewards()


class gym_reward_lambda(gym.Wrapper):
    def __init__(self, env, change_reward_fn):
        assert callable(change_reward_fn), "change_reward_fn needs to be a function. It is {}".format(change_reward_fn)
        self._change_reward_fn = change_reward_fn

        super().__init__(env)

    def step(self, action):
        obs, rew, done, info = super().step(action)
        return obs, self._change_reward_fn(rew), done, info


reward_lambda_v0 = WrapperChooser(aec_wrapper=aec_reward_lambda, gym_wrapper=gym_reward_lambda)



from .utils.base_modifier import BaseModifier
from .utils.shared_wrapper_util import shared_wrapper
from supersuit.utils.accumulator import Accumulator
import numpy as np


def max_observation_v0(env, memory):
    int(memory)  # delay must be an int

    class MaxObsModifier(BaseModifier):
        def reset(self):
            self.accumulator = Accumulator(self.observation_space, memory, np.maximum)

        def modify_obs(self, obs):
            self.accumulator.add(obs)
            return super().modify_obs(self.accumulator.get())

    return shared_wrapper(env, MaxObsModifier)



from supersuit.utils import agent_indicator as agent_ider
from pettingzoo.utils.env import AECEnv, ParallelEnv
from supersuit import observation_lambda_v0


def agent_indicator_v0(env, type_only=False):
    assert isinstance(env, AECEnv) or isinstance(env, ParallelEnv), "agent_indicator_v0 only accepts an AECEnv or ParallelEnv"

    indicator_map = agent_ider.get_indicator_map(env.possible_agents, type_only)
    num_indicators = len(set(indicator_map.values()))

    agent_ider.check_params(env.observation_spaces.values())

    return observation_lambda_v0(env,
        lambda obs, obs_space, agent: agent_ider.change_observation(
            obs,
            obs_space,
            (indicator_map[agent], num_indicators),
        ),
        lambda obs_space: agent_ider.change_obs_space(obs_space, num_indicators))



from .async_vector_env import AsyncAECVectorEnv
from .vector_env import SyncAECVectorEnv
import cloudpickle


def vectorize_aec_env_v0(aec_env, num_envs, num_cpus=0):
    def env_fn():
        return cloudpickle.loads(cloudpickle.dumps(aec_env))

    env_list = [env_fn] * num_envs

    if num_cpus == 0 or num_cpus == 1:
        return SyncAECVectorEnv(env_list)
    else:
        return AsyncAECVectorEnv(env_list, num_cpus)



import numpy as np
import gym.vector


class MarkovVectorEnv(gym.vector.VectorEnv):
    def __init__(self, par_env, black_death=False):
        """
        parameters:
            - par_env: the pettingzoo Parallel environment that will be converted to a gym vector environment
            - black_death: whether to give zero valued observations and 0 rewards when an agent is done, allowing for environments with multiple numbers of agents.
                            Is equivalent to adding the black death wrapper, but somewhat more efficient.

        The resulting object will be a valid vector environment that has a num_envs
        parameter equal to the max number of agents, will return an array of observations,
        rewards, dones, etc, and will reset environment automatically when it finishes
        """
        self.par_env = par_env
        self.metadata = par_env.metadata
        self.observation_space = list(par_env.observation_spaces.values())[0]
        self.action_space = list(par_env.action_spaces.values())[0]
        assert all(
            self.observation_space == obs_space for obs_space in par_env.observation_spaces.values()
        ), "observation spaces not consistent. Perhaps you should wrap with `supersuit.aec_wrappers.pad_observations`?"
        assert all(
            self.action_space == obs_space for obs_space in par_env.action_spaces.values()
        ), "action spaces not consistent. Perhaps you should wrap with `supersuit.aec_wrappers.pad_actions`?"
        self.num_envs = len(par_env.possible_agents)
        self.black_death = black_death
        self.obs_buffer = np.empty((self.num_envs,) + self.observation_space.shape, dtype=self.observation_space.dtype)

    def seed(self, seed=None):
        self.par_env.seed(seed)

    def concat_obs(self, obs_dict):
        if self.black_death:
            self.obs_buffer[:] = 0
        for i, agent in enumerate(self.par_env.possible_agents):
            if self.black_death:
                if agent in obs_dict:
                    self.obs_buffer[i] = obs_dict[agent]
            else:
                if agent not in obs_dict:
                    raise AssertionError("environment has agent death. Not allowed for pettingzoo_env_to_vec_env_v0 unless black_death is True")
                self.obs_buffer[i] = obs_dict[agent]

        return self.obs_buffer

    def step_async(self, actions):
        self._saved_actions = actions

    def step_wait(self):
        return self.step(self._saved_actions)

    def reset(self):
        return self.concat_obs(self.par_env.reset())

    def step(self, actions):
        agent_set = set(self.par_env.agents)
        act_dict = {agent: actions[i] for i, agent in enumerate(self.par_env.possible_agents) if agent in agent_set}
        observations, rewards, dones, infos = self.par_env.step(act_dict)

        # adds last observation to info where user can get it
        if all(dones.values()):
            for agent, obs in observations.items():
                infos[agent]['terminal_observation'] = obs

        rews = np.array([rewards.get(agent, 0) for agent in self.par_env.possible_agents], dtype=np.float32)
        dns = np.array([dones.get(agent, False) for agent in self.par_env.possible_agents], dtype=np.uint8)
        infs = [infos.get(agent, {}) for agent in self.par_env.possible_agents]

        if all(dones.values()):
            observations = self.reset()
        else:
            observations = self.concat_obs(observations)
        assert (
            self.black_death or self.par_env.agents == self.par_env.possible_agents
        ), "MarkovVectorEnv does not support environments with varying numbers of active agents unless black_death is set to True"
        return observations, rews, dns, infs

    def render(self, mode="human"):
        return self.par_env.render(mode)

    def close(self):
        return self.par_env.close()

    def env_is_wrapped(self, wrapper_class):
        '''
        env_is_wrapped only suppors vector and gym environments
        currently, not pettingzoo environments
        '''
        return [False] * self.num_envs



import gym
from pettingzoo.utils.env import AECEnv, ParallelEnv
from pettingzoo.utils.conversions import to_parallel, from_parallel


class WrapperChooser:
    def __init__(self, aec_wrapper=None, gym_wrapper=None, parallel_wrapper=None):
        assert aec_wrapper is not None or parallel_wrapper is not None, "either the aec wrapper or the parallel wrapper must be defined for all supersuit environments"
        self.aec_wrapper = aec_wrapper
        self.gym_wrapper = gym_wrapper
        self.parallel_wrapper = parallel_wrapper

    def __call__(self, env, *args, **kwargs):
        if isinstance(env, gym.Env):
            if self.gym_wrapper is None:
                raise ValueError(f"{self.wrapper_name} does not apply to gym environments, pettingzoo environments only")
            return self.gym_wrapper(env, *args, **kwargs)
        elif isinstance(env, AECEnv):
            if self.aec_wrapper is not None:
                return self.aec_wrapper(env, *args, **kwargs)
            else:
                return from_parallel(self.parallel_wrapper(to_parallel(env), *args, **kwargs))
        elif isinstance(env, ParallelEnv):
            if self.parallel_wrapper is not None:
                return self.parallel_wrapper(env, *args, **kwargs)
            else:
                return to_parallel(self.aec_wrapper(from_parallel(env), *args, **kwargs))
        else:
            raise ValueError("environment passed to supersuit wrapper must either be a gym environment or a pettingzoo environment")
